/*
 *  aplay_engine.c	(C) 1999 by Andy Lo A foe
 *  Based on aplay Copyright (c) by Jaroslav Kysela <perex@jcu.cz>
 *  Based on vplay program by Michael Beck
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <malloc.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <netinet/in.h>
#include <glib.h>
#include "formats.h"
#include "../SampleBuffer.h"
#include "../input_plugin.h"

extern SampleBuffer *write_buffer;

#define APLAY_FRAME_SIZE	4608

#define FORMAT_DEFAULT		-1
#define FORMAT_RAW		0
#define FORMAT_VOC		1
#define FORMAT_WAVE		2
#define FORMAT_AU		3

/* global data */

static char wav_name[FILENAME_MAX+1];
static int count;

static FILE *wav_fd;
static int header_size;

typedef struct _the_format
{
	int rate;
	int channels;
	int bits;
} the_format;

static the_format format;


void init_aplay()
{
	format.rate = 0;
	format.channels = 0;
	format.bits = 0;
	return;
}


/*
 * test, if it's a .WAV file, 0 if ok (and set the speed, stereo etc.)
 *                            < 0 if not
 */
static int test_wavefile(void *buffer)
{
	WaveHeader *wp = (WaveHeader *)buffer;

	if (wp->main_chunk == WAV_RIFF && wp->chunk_type == WAV_WAVE &&
	    wp->sub_chunk == WAV_FMT && wp->data_chunk == WAV_DATA) {
		if (wp->format != WAV_PCM_CODE) {
			fprintf(stderr, "APLAY: can't play not PCM-coded WAVE-files\n");
			return -1;
		}
		if (wp->modus < 1 || wp->modus > 32) {
			fprintf(stderr, "APLAY: can't play WAVE-files with %d tracks\n",
				wp->modus);
			return -1;
		}
		switch (wp->bit_p_spl) {
		case 8:
			//fprintf(stderr, "APLAY: no support for 8-bit  yet\n");
			//return -1;
		case 16:
			// Humm, no nothing :)
			break;
		default:
			fprintf(stderr, "APLAY: can't play WAVE-files with sample %d bits wide\n",
				wp->bit_p_spl);
		}
		format.rate = wp->sample_fq;
		format.channels = wp->modus;
		format.bits = wp->bit_p_spl;
		count = wp->data_length;
		return 0;
	}
	return -1;
}


int aplay_open(input_object *obj, char *name)
{
	char audiobuf[4096];

	if (!name || !strcmp(name, "-")) {
		printf("APLAY: Uhm, we don't support stdin\n");
		return 0;
	} else {
		if ((wav_fd = fopen(name, "r")) == NULL) {
			perror(name);
			return 0;
		}
	}
	init_aplay();

	/* read the file header */
	if (fread(audiobuf, sizeof(WaveHeader), 1, wav_fd) != 1) {
		fprintf(stderr, "APLAY: read error");
		fclose(wav_fd);
		return 0;
	}
	if (test_wavefile(audiobuf) >= 0) {
		// Extract the filename
		char *ptr = strrchr(name, '/');
		if (ptr)
			ptr++;
		else
			ptr = name;
		if (strlen(ptr) > FILENAME_MAX) {
			strncpy(wav_name, ptr, FILENAME_MAX-1);
			wav_name[FILENAME_MAX-1] = 0;
		} else 
			strcpy(wav_name, ptr);
		header_size = sizeof(WaveHeader);	
		return 1;
	}
	if (wav_fd != NULL)
		fclose(wav_fd);
	return 0;
}


void aplay_close(input_object *obj)
{
	if (wav_fd != NULL)
			fclose(wav_fd);
	wav_fd = 0;
}

int aplay_play_frame(input_object *obj)
{
	char audiobuf[8192];
	char tmpbuf[8192];
	
	// Take care of mono streams
	if (format.channels == 1) { // mono, so double 
			if (format.bits == 8) { // 8 bit
				if (fread(tmpbuf, APLAY_FRAME_SIZE >> 2, 1, wav_fd) != 1) {
					return 0;
				}
				guint8 *s = (guint8 *)tmpbuf;
				gint16 *d = (gint16 *)audiobuf;
				gint16 c;
				for (int i=0; i < APLAY_FRAME_SIZE; i+=4) {
					*d++ = c = ((*s ^ 0x80) << 8) | *s++;
					*d++ = c;
				}
			} else { // 16 bit					
				if (fread(tmpbuf, APLAY_FRAME_SIZE >> 1, 1, wav_fd) != 1) {
						return 0;
				}
				gint16 *s = (gint16 *)tmpbuf;
				gint16 *d = (gint16 *)audiobuf;
				for (int i=0; i < APLAY_FRAME_SIZE; i+=4) {
						*d++ = *s;
						*d++ = *s++; // Copy twice
				}
			}
	} else if (format.channels == 2) {
		if (fread(audiobuf, APLAY_FRAME_SIZE, 1, wav_fd) != 1) {
				return 0;
		}
	} else {
			printf("Huh? More than 2 channels?\n");
			exit(3);
	}
	if (write_buffer) {
			write_buffer->WriteSamples(audiobuf, APLAY_FRAME_SIZE >> 2);
	}
	return 1;
}


int aplay_seek(input_object *obj, int frame)
{
	long position = (frame * (APLAY_FRAME_SIZE >> (2 - format.channels)));
	position += header_size;

	return !fseek(wav_fd, position, SEEK_SET);
}


int aplay_frame_size(input_object *obj)
{
	return APLAY_FRAME_SIZE;
}


int aplay_nr_frames(input_object *obj)
{
	if (count)
			return (count / (APLAY_FRAME_SIZE >> ((2 - format.channels) + (format.bits == 8 ? 1 : 0))));
	else
			return 0;
}
		

int aplay_sample_rate(input_object *obj)
{
	return format.rate;
}


int aplay_channels(input_object *obj)
{
	return 2; // Yes, always stereo, we take care of mono -> stereo converion
}


unsigned long aplay_frame_to_sec(input_object *obj, int frame)
{
	unsigned long pos = frame * (APLAY_FRAME_SIZE >> (2 - format.channels));

	return (pos / (format.rate / 100) / format.channels / 2);
}



float aplay_test_support(char *name)
{
	WaveHeader wp;
	FILE *fd;

	if ((fd = fopen(name, "r")) == NULL) {
		return 0.0;
	}	
	if (fread(&wp, sizeof(WaveHeader), 1, fd) != 1) {
			fclose(fd);
			return 0.0;
	}
	fclose(fd);
    if (wp.main_chunk == WAV_RIFF && wp.chunk_type == WAV_WAVE &&
        wp.sub_chunk == WAV_FMT && wp.data_chunk == WAV_DATA) {
		if (wp.format != WAV_PCM_CODE) {
			fprintf(stderr, "APLAY: can't play not PCM-coded WAVE-files\n");
			return 0.0;
		}
	}
	return 1.0;
}

int aplay_stream_info(input_object *obj, struct stream_info *info)
{
	if (!info)
			return 0;
	sprintf(info->stream_type, "%d-bit %dKhz %s WAV", 16, 
		aplay_sample_rate(obj) / 1000, aplay_channels(obj) == 2 ? "stereo" : "mono");
	info->author[0] = 0;
	strcpy(info->title, wav_name);	

	return 1;
}

int aplay_init()
{
	return 1;
}


input_plugin aplay_plugin = {
	{ "aplay WAV player  v1.0" },			 // Plugin name
	{ "Jaroslav Kysela and Andy Lo A Foe" }, // Author name
	aplay_init,
	aplay_test_support,
	aplay_open,
	aplay_close,
	aplay_play_frame,
	aplay_seek,
	aplay_frame_size,
	aplay_nr_frames,
	aplay_frame_to_sec,
	aplay_sample_rate,
	aplay_channels,
	aplay_stream_info	
};
